(******************************************************************************)
(*                                                                            *)
(*                                    Fix                                     *)
(*                                                                            *)
(*                       François Pottier, Inria Paris                        *)
(*                                                                            *)
(*  Copyright Inria. All rights reserved. This file is distributed under the  *)
(*  terms of the GNU Library General Public License version 2, with a         *)
(*  special exception on linking, as described in the file LICENSE.           *)
(*                                                                            *)
(******************************************************************************)

(**The Boolean lattice. The ordering is [false <= true]. *)

open Sigs

include PROPERTY with type property = bool
include SEMI_LATTICE with type property := bool
include MINIMAL_SEMI_LATTICE with type property := bool
